"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const cli_1 = require("../../cli");
const session_1 = require("../../session");
const chalk_1 = require("chalk");
const shelljs = require("shelljs");
const hassConfiguration_1 = require("../../models/hassConfiguration");
const utils_1 = require("../../utils");
const ini = require("ini");
const path = require("path");
exports.command = 'skills';
exports.desc = 'Install skill present on the device';
exports.builder = {
    git: {
        demandOption: false,
        describe: 'Install a skill from a git repository',
        type: 'string',
        alias: 'g',
    },
    branch: {
        demandOption: false,
        describe: 'Specify the git branch you want to use',
        type: 'string',
        alias: 'b',
    },
};
exports.handler = (argv) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const git = JSON.parse(JSON.stringify(argv)).git;
    const branch = JSON.parse(JSON.stringify(argv)).branch;
    const ssh = new session_1.SSHService();
    yield ssh.connect()
        .catch(e => {
        cli_1.cli.stream.error(e);
        process.exit();
    });
    cli_1.cli.stream.println(chalk_1.default.green('Checking for skills'));
    const snipsfile = yield ssh.getAssistantSnipsfile()
        .catch(e => cli_1.cli.stream.error('Error reading the Snipsfile for skills, reason: ' + e.message));
    const skills = yield ssh.getSnippets()
        .then(snippets => {
        const skillCount = snippets.size;
        cli_1.cli.stream.success();
        if (skillCount === 1) {
            cli_1.cli.stream.hint('Found 1 skill');
        }
        else {
            cli_1.cli.stream.hint(`Found ${skillCount} skills`);
        }
        return snippets;
    })
        .catch(_ => { });
    if (skills === undefined)
        cli_1.cli.stream.println('No python or Home Assistant snippets found');
    const skillSetup = new SkillSetup(ssh, skills, snipsfile);
    yield ssh.dpkgCheckIfPackageIsInstalled(session_1.SnipsServices.template)
        .then(result => {
        if (!result) {
            cli_1.cli.stream.loading('Installing ' + session_1.SnipsServices.template + ' and ' + session_1.SnipsServices.skillServer);
            return ssh.run(`sudo apt-get install -y ${session_1.SnipsServices.template} ${session_1.SnipsServices.skillServer}`)
                .then(_ => cli_1.cli.stream.success())
                .then(_ => 'success')
                .catch(e => {
                cli_1.cli.stream.error(`Can't install packages ${session_1.SnipsServices.template} and ${session_1.SnipsServices.skillServer}, error : ${e.message}`);
                cli_1.cli.stream.hint('Your device architecture might not be supported by those two packages, please contact us');
                ssh.disconnect();
                process.exit(0);
            });
        }
        return 'success';
    })
        .then(_ => ssh.addCurrentUserToSnipsSkillsAdminGroup().catch(_ => { }));
    yield ssh.stopService(session_1.SnipsServices.skillServer);
    if (skillSetup.hasGitSkill || (git !== undefined && git.length > 0)) {
        if (!(yield ssh.checkCommandExists('virtualenv'))) {
            cli_1.cli.stream.loading('Installing pip package and virtualenv. This could take a while.');
            yield ssh.installPythonAndPip(_ => { });
            cli_1.cli.stream.success();
        }
        if (!(yield ssh.checkCommandExists('git'))) {
            cli_1.cli.stream.loading('Installing git package');
            yield ssh.aptGetY(['git'], _ => { });
            cli_1.cli.stream.success();
        }
        if (snipsfile !== undefined && snipsfile.skills.length > 0) {
            for (const skill of snipsfile.skills) {
                if (skill.url) {
                    cli_1.cli.stream.println(`Cloning skill ${skill.name} from url: ${skill.url}`);
                    yield ssh.gitClone(skill.url, undefined, output => cli_1.cli.stream.println(output));
                }
            }
        }
        if (git !== undefined && git.length > 0) {
            cli_1.cli.stream.println(`Cloning skill from url: ${git}`);
            yield ssh.gitClone(git, branch, output => cli_1.cli.stream.println(output));
        }
    }
    cli_1.cli.stream.println(chalk_1.default.green('Generate skills based on the snippets provided'));
    yield ssh.snipsTemplateRender(output => cli_1.cli.stream.print(output));
    if (skills !== undefined) {
        if (skillSetup.hasPythonSkill) {
            cli_1.cli.stream.println(chalk_1.default.green('\nSetting up Python skills'));
            yield ssh.checkCommandExists('virtualenv')
                .then(isVirtualEnvInstalled => {
                if (!isVirtualEnvInstalled) {
                    cli_1.cli.stream.loading('Installing pip and virtualenv. This could take a while.');
                    return ssh.installPythonAndPip(_ => { })
                        .then(_ => cli_1.cli.stream.success());
                }
                return;
            })
                .catch(e => cli_1.cli.stream.error('Unable to install the python skill, reason: ' + e.message));
        }
        if (skillSetup.hasHassSkill) {
            cli_1.cli.stream.println(chalk_1.default.green('\nSetting up Home Assistant skills'));
            yield skillSetup.setupHomeAssistantSkill()
                .then(_ => cli_1.cli.stream.loading('Relaunching Home Assistant'))
                .then(_ => ssh.restartHomeAssistant())
                .then(_ => cli_1.cli.stream.success())
                .catch(e => cli_1.cli.stream.error('Unable to install Home Assistant skill, reason: ' + e.message));
        }
    }
    try {
        cli_1.cli.stream.println(chalk_1.default.green('\nChecking skills for end-user parameters'));
        yield skillSetup.promptForSecretParameters();
        cli_1.cli.stream.success();
    }
    catch (error) {
        cli_1.cli.stream.error(error.message);
    }
    if (skillSetup.hasPythonSkill || skillSetup.hasGitSkill) {
        try {
            try {
                cli_1.cli.stream.loading('Generating virtual environment for python skills');
                yield ssh.runSkillSetupSH();
                cli_1.cli.stream.success();
            }
            catch (error) {
                throw new Error('Error setting up virtualenv, reason : ' + error.message);
            }
            cli_1.cli.stream.println('Relaunching snips-skill-server');
            yield ssh.relaunchService(session_1.SnipsServices.skillServer);
        }
        catch (error) {
            cli_1.cli.stream.error(error.message);
        }
    }
    const assistant = yield ssh.getInstalledAssistant();
    if (assistant !== undefined) {
        cli_1.cli.stream.success(`Snips assistant is now running. Say ${chalk_1.default.blue(assistant.hotword)} to start!`);
    }
    else {
        cli_1.cli.stream.success(`Snips assistant is now running`);
    }
    ssh.disconnect();
    process.exit();
});
class SkillSetup {
    constructor(ssh, snippets, snipsfile) {
        this.ssh = ssh;
        this.snippets = snippets;
        this.snipsfile = snipsfile;
        this.hassSnippets = new Map();
        this.hassCfgPath = '/home/homeassistant/.homeassistant/';
        snippets && snippets.forEach((skill, key) => {
            if (skill.type.includes('python')) {
                this.hasPythonSkill = true;
            }
            if (skill.type.includes('homeassistant')) {
                this.hasHassSkill = true;
                this.hassSnippets.set(key, skill.snippets);
            }
        });
        this.hasGitSkill = snipsfile !== undefined && snipsfile.skills.length > 0;
    }
    updateConfigSecrets(folder, skillName = folder) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const rawConfig = yield this.ssh.getSkillConfig(folder);
            const config = rawConfig && ini.parse(rawConfig);
            const configPath = `/var/lib/snips/skills/${folder}/config.ini`;
            if (config && config.secret && Object.keys(config.secret).length > 0) {
                cli_1.cli.stream.hint(`Skill ${skillName} contains end-user parameters that need to be configured`);
                const secrets = Object.entries(config.secret);
                const answers = yield cli_1.cli.prompt.promptQuestions(secrets.map(([key, val]) => ({
                    name: key,
                    default: val,
                    message: `Please enter a value for the parameter [${key}]`,
                })));
                Object.entries(answers).forEach(([key, value]) => {
                    config.secret[key] = value;
                });
                cli_1.cli.stream.hint(`You can change these values later by editing the file located at ${configPath}`);
                const newConfig = ini.stringify(config);
                yield this.ssh.updateSkillConfig(newConfig, folder);
            }
        });
    }
    promptForSecretParameters() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this.snippets) {
                for (const [key, value] of this.snippets) {
                    yield this.updateConfigSecrets(key);
                }
            }
            if (this.snipsfile) {
                for (const skill of this.snipsfile.skills) {
                    if (skill.url) {
                        const folder = path.basename(skill.url).replace(/\.[^/.]+$/, '');
                        yield this.updateConfigSecrets(folder, skill.name);
                    }
                }
            }
        });
    }
    setupHomeAssistantSkill() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.ssh.fileExists(this.hassCfgPath + 'configuration.yaml')
                .then(exists => {
                if (!exists)
                    return this.askUserForCustomHassInstallConfigurationPath();
                cli_1.cli.stream.success('Found Home Assistant configuration.yaml');
                return this.hassCfgPath;
            })
                .then(cfgFilePath => {
                if (cfgFilePath === undefined)
                    return this.installHass();
                return cfgFilePath;
            })
                .then(cfgFilePath => this.hassCfgPath = cfgFilePath.replace(/\/?$/, '/'))
                .then(_ => cli_1.cli.stream.println('Moving generated skills to Hass python_script folder'))
                .then(_ => this.ssh.moveSkillsToHomeAssistantFolder(this.hassCfgPath + 'python_scripts', Array.from(this.hassSnippets.keys())))
                .then(_ => cli_1.cli.stream.println('Backing up your last Hass configuration'))
                .then(_ => this.ssh.backupHassConfiguration(this.hassCfgPath))
                .then(_ => this.ssh.getFile(this.hassCfgPath + 'configuration.yaml'))
                .then(file => {
                const parsedYamlFile = hassConfiguration_1.HassConfiguration.parse(file);
                if (parsedYamlFile === undefined)
                    throw new Error('Error parsing the Home Assistant configuration');
                return parsedYamlFile;
            })
                .then(hassCfgFile => {
                cli_1.cli.stream.println('Creating new Hass configuration');
                return this.ssh.getInstalledAssistant()
                    .then(assistantFile => {
                    return hassConfiguration_1.HassConfiguration.generateSnipsConfiguration(hassCfgFile, assistantFile, this.hassSnippets);
                });
            })
                .then(newYamlFilePath => {
                return this.ssh.putFile(newYamlFilePath, '/tmp/configuration.yaml')
                    .then(_ => shelljs.rm(newYamlFilePath))
                    .then(_ => this.ssh.run(`sudo mv /tmp/configuration.yaml ${this.hassCfgPath}`));
            });
        });
    }
    askUserForCustomHassInstallConfigurationPath() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return cli_1.cli.prompt.promptQuestions([
                {
                    name: 'isInstalled',
                    type: 'confirm',
                    message: 'Is Home Assistant already installed on this device?',
                },
            ])
                .then(isHassInstalled => {
                if (isHassInstalled['isInstalled']) {
                    let hassCfgFound = false;
                    return cli_1.cli.prompt.promptQuestions([
                        {
                            name: 'hassCfgPath',
                            type: 'input',
                            message: 'Enter the path to Home Assistant configuration file.',
                            validate: (value) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                                if (value.length === 0)
                                    return `Path can't be blank`;
                                const pathWithSlash = value.replace(/\/?$/, '/');
                                hassCfgFound = yield this.ssh.fileExists(pathWithSlash + 'configuration.yaml');
                                if (!hassCfgFound)
                                    return `Didn't find ${pathWithSlash}configuration.yaml on the device`;
                                return true;
                            }),
                        },
                    ])
                        .then(cfgPathResponse => cfgPathResponse['hassCfgPath']);
                }
                return undefined;
            });
        });
    }
    installHass() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const hassInstallation = yield cli_1.cli.prompt.promptQuestions([
                {
                    name: 'installHass',
                    type: 'confirm',
                    message: 'You need to install Home Assistant to continue, proceed?',
                },
            ]);
            if (!hassInstallation['installHass']) {
                cli_1.cli.stream.println(`Quitting`);
                cli_1.cli.stream.hint('You can install Home Assistant by running ' + chalk_1.default.blue('sam hass install'));
                throw new Error(`Hass isn't installed, aborting`);
            }
            cli_1.cli.stream.println('Installing Home Assistant, this can take a while');
            return yield this.ssh.installHomeAssistant(_ => { }, stepTitle => {
                cli_1.cli.stream.println(chalk_1.default.green(stepTitle));
            })
                .then((_) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                let loadingText = 'Waiting for Home Assistant initial setup. This could take a while.';
                cli_1.cli.stream.loading(loadingText);
                let homeAssistantConfigurationDetected = false;
                while (homeAssistantConfigurationDetected === false) {
                    yield utils_1.timeout(5000);
                    homeAssistantConfigurationDetected = yield this.ssh.fileExists('/home/homeassistant/.homeassistant/configuration.yaml');
                    loadingText += ' . ';
                    cli_1.cli.stream.loading(loadingText);
                }
                cli_1.cli.stream.success();
                return 'success';
            }))
                .then(_ => this.hassCfgPath);
        });
    }
}
