import * as yargs from 'yargs'
import { cli } from '../../cli'
import { SSHService, SnipsServices } from '../../session'
import chalk from 'chalk'
import * as shelljs from 'shelljs'
import { HassConfiguration } from '../../models/hassConfiguration'
import { timeout } from '../../utils'
import { Snipsfile } from '../../models'
import * as ini from 'ini'
import path = require('path')

exports.command = 'skills'
exports.desc = 'Install skill present on the device'

exports.builder = {
    git: {
        demandOption: false,
        describe: 'Install a skill from a git repository',
        type: 'string',
        alias: 'g',
    },

    branch: {
        demandOption: false,
        describe: 'Specify the git branch you want to use',
        type: 'string',
        alias: 'b',
    },
}

exports.handler = async (argv: yargs.Argv) => {
    const git: string = JSON.parse(JSON.stringify(argv)).git
    const branch: string = JSON.parse(JSON.stringify(argv)).branch

    const ssh = new SSHService()
    await ssh.connect()
    .catch(e => {
        cli.stream.error(e)
        process.exit()
    })

    cli.stream.println(chalk.green('Checking for skills'))
    const snipsfile = await ssh.getAssistantSnipsfile()
    .catch(e => cli.stream.error('Error reading the Snipsfile for skills, reason: ' + e.message))

    const skills = await ssh.getSnippets()
    .then(snippets => {
        const skillCount = snippets.size
        cli.stream.success()
        if (skillCount === 1) {
            cli.stream.hint('Found 1 skill')
        } else {
            cli.stream.hint(`Found ${skillCount} skills`)
        }
        return snippets
    })
    .catch(_ => {})

    if (skills === undefined) cli.stream.println('No python or Home Assistant snippets found')

    const skillSetup = new SkillSetup(ssh, skills, snipsfile)

    await ssh.dpkgCheckIfPackageIsInstalled(SnipsServices.template)
    .then(result => {
        if (!result) {
            cli.stream.loading('Installing ' + SnipsServices.template + ' and ' + SnipsServices.skillServer)
            return ssh.run(`sudo apt-get install -y ${SnipsServices.template} ${SnipsServices.skillServer}`)
            .then(_ => cli.stream.success())
            .then(_ => 'success')
            .catch(e => {
                cli.stream.error(`Can't install packages ${SnipsServices.template} and ${SnipsServices.skillServer}, error : ${e.message}`)
                cli.stream.hint('Your device architecture might not be supported by those two packages, please contact us')
                ssh.disconnect()
                process.exit(0)
            })
        }
        return 'success'
    })
    .then(_ => ssh.addCurrentUserToSnipsSkillsAdminGroup().catch(_ => {}))

    await ssh.stopService(SnipsServices.skillServer)

    if (skillSetup.hasGitSkill || (git !== undefined && git.length > 0)) {
        if (!await ssh.checkCommandExists('virtualenv')) {
            cli.stream.loading('Installing pip package and virtualenv. This could take a while.')
            await ssh.installPythonAndPip(_ => {})
            cli.stream.success()
        }
        if (!await ssh.checkCommandExists('git')) {
            cli.stream.loading('Installing git package')
            await ssh.aptGetY(['git'], _ => {})
            cli.stream.success()
        }
        if (snipsfile !== undefined && snipsfile.skills.length > 0) {
            for (const skill of snipsfile.skills) {
                if (skill.url) {
                    cli.stream.println(`Cloning skill ${skill.name} from url: ${skill.url}`)
                    await ssh.gitClone(skill.url, undefined, output => cli.stream.println(output))
                }
            }
        }
        if (git !== undefined && git.length > 0) {
            cli.stream.println(`Cloning skill from url: ${git}`)
            await ssh.gitClone(git, branch, output => cli.stream.println(output))
        }
    }

    cli.stream.println(chalk.green('Generate skills based on the snippets provided'))
    await ssh.snipsTemplateRender(output => cli.stream.print(output))

    if (skills !== undefined) {
        if (skillSetup.hasPythonSkill) {
            cli.stream.println(chalk.green('\nSetting up Python skills'))
            await ssh.checkCommandExists('virtualenv')
            .then(isVirtualEnvInstalled => {
                if (!isVirtualEnvInstalled) {
                    cli.stream.loading('Installing pip and virtualenv. This could take a while.')
                    return ssh.installPythonAndPip(_ => {})
                    .then(_ => cli.stream.success())
                }
                return
            })
            .catch(e => cli.stream.error('Unable to install the python skill, reason: ' + e.message))
        }

        if (skillSetup.hasHassSkill) {
            cli.stream.println(chalk.green('\nSetting up Home Assistant skills'))
            await skillSetup.setupHomeAssistantSkill()
            .then(_ => cli.stream.loading('Relaunching Home Assistant'))
            .then(_ => ssh.restartHomeAssistant())
            .then(_ => cli.stream.success())
            .catch(e => cli.stream.error('Unable to install Home Assistant skill, reason: ' + e.message))
        }
    }

    try {
        cli.stream.println(chalk.green('\nChecking skills for end-user parameters'))
        await skillSetup.promptForSecretParameters()
        cli.stream.success()
    } catch (error) {
        cli.stream.error(error.message)
    }

    if (skillSetup.hasPythonSkill || skillSetup.hasGitSkill) {
        try {
            try {
                cli.stream.loading('Generating virtual environment for python skills')
                await ssh.runSkillSetupSH()
                cli.stream.success()
            } catch (error) {
                throw new Error('Error setting up virtualenv, reason : ' + error.message)
            }
            cli.stream.println('Relaunching snips-skill-server')
            await ssh.relaunchService(SnipsServices.skillServer)
        } catch (error) {
            cli.stream.error(error.message)
        }
    }

    const assistant = await ssh.getInstalledAssistant()
    if (assistant !== undefined) {
        cli.stream.success(`Snips assistant is now running. Say ${chalk.blue(assistant.hotword)} to start!`)
    } else {
        cli.stream.success(`Snips assistant is now running`)
    }

    ssh.disconnect()
    process.exit()
}

type SkillSnippet = { type: string, snippets: string[] }

class SkillSetup {


    hasPythonSkill: boolean
    hasHassSkill: boolean
    hasGitSkill: boolean

    hassSnippets: Map<string, string[]> = new Map<string, string[]>()

    hassCfgPath = '/home/homeassistant/.homeassistant/'

    constructor(
        private ssh: SSHService,
        private snippets: void | Map<string, SkillSnippet>,
        private snipsfile: void | Snipsfile) {

        snippets && snippets.forEach((skill, key) => {
            if (skill.type.includes('python')) {
                this.hasPythonSkill = true
            }
            if (skill.type.includes('homeassistant')) {
                this.hasHassSkill = true
                this.hassSnippets.set(key, skill.snippets)
            }
        })
        this.hasGitSkill = snipsfile !== undefined && snipsfile.skills.length > 0
    }

    async updateConfigSecrets (folder, skillName = folder) {
        const rawConfig = await this.ssh.getSkillConfig(folder)
        const config = rawConfig && ini.parse(rawConfig)
        const configPath = `/var/lib/snips/skills/${folder}/config.ini`
        if (config && config.secret && Object.keys(config.secret).length > 0) {
            cli.stream.hint(`Skill ${skillName} contains end-user parameters that need to be configured`)
            const secrets = Object.entries(config.secret)
            const answers = await cli.prompt.promptQuestions(secrets.map(([key, val]) => ({
                name: key,
                default: val,
                message: `Please enter a value for the parameter [${key}]`,
            })))
            Object.entries(answers).forEach(([key, value]) => {
                config.secret[key] = value
            })
            cli.stream.hint(`You can change these values later by editing the file located at ${configPath}`)
            const newConfig = ini.stringify(config)
            await this.ssh.updateSkillConfig(newConfig, folder)
        }
    }

    async promptForSecretParameters () {
        if (this.snippets) {
            for (const [key, value] of this.snippets) {
                await this.updateConfigSecrets(key)
            }
        }
        if (this.snipsfile) {
            for (const skill of this.snipsfile.skills) {
                if (skill.url) {
                    const folder = path.basename(skill.url).replace(/\.[^/.]+$/, '')
                    await this.updateConfigSecrets(folder, skill.name)
                }
            }
        }
    }

    async setupHomeAssistantSkill() {
        // Check for existing Hass configuration file
        await this.ssh.fileExists(this.hassCfgPath + 'configuration.yaml')
        .then(exists => {
            if (!exists) return this.askUserForCustomHassInstallConfigurationPath()
            cli.stream.success('Found Home Assistant configuration.yaml')
            return this.hassCfgPath
        })
        .then(cfgFilePath => {
            if (cfgFilePath === undefined) return this.installHass()
            return cfgFilePath
        })
        .then(cfgFilePath => this.hassCfgPath = cfgFilePath.replace(/\/?$/, '/'))
        .then(_ => cli.stream.println('Moving generated skills to Hass python_script folder'))
        .then(_ => this.ssh.moveSkillsToHomeAssistantFolder(this.hassCfgPath + 'python_scripts', Array.from(this.hassSnippets.keys())))
        .then(_ => cli.stream.println('Backing up your last Hass configuration'))
        .then(_ => this.ssh.backupHassConfiguration(this.hassCfgPath))
        .then(_ => this.ssh.getFile(this.hassCfgPath + 'configuration.yaml'))
        .then(file => {
            const parsedYamlFile = HassConfiguration.parse(file)
            if (parsedYamlFile === undefined) throw new Error('Error parsing the Home Assistant configuration')
            return parsedYamlFile
        })
        .then(hassCfgFile => {
            cli.stream.println('Creating new Hass configuration')
            return this.ssh.getInstalledAssistant()
            .then(assistantFile => {
                return HassConfiguration.generateSnipsConfiguration(hassCfgFile, assistantFile, this.hassSnippets)
            })
        })
        .then(newYamlFilePath => {
            return this.ssh.putFile(newYamlFilePath, '/tmp/configuration.yaml')
            .then(_ => shelljs.rm(newYamlFilePath))
            .then(_ => this.ssh.run(`sudo mv /tmp/configuration.yaml ${this.hassCfgPath}`))
        })
    }

    private async askUserForCustomHassInstallConfigurationPath(): Promise<string | undefined> {
        return cli.prompt.promptQuestions([
            {
                name: 'isInstalled',
                type: 'confirm',
                message: 'Is Home Assistant already installed on this device?',
            },
        ])
        .then(isHassInstalled => {
            if (isHassInstalled['isInstalled']) {
                let hassCfgFound = false
                return cli.prompt.promptQuestions([
                    {
                        name: 'hassCfgPath',
                        type: 'input',
                        message: 'Enter the path to Home Assistant configuration file.',
                        validate: async value => {
                            if (value.length === 0) return `Path can't be blank`
                            const pathWithSlash = value.replace(/\/?$/, '/')
                            hassCfgFound = await this.ssh.fileExists(pathWithSlash + 'configuration.yaml')
                            if (!hassCfgFound) return `Didn't find ${pathWithSlash}configuration.yaml on the device`
                            return true
                        },
                    },
                ])
                .then(cfgPathResponse => cfgPathResponse['hassCfgPath'])
            }
            return undefined
        })
    }

    private async installHass(): Promise<string> {
        const hassInstallation = await cli.prompt.promptQuestions([
            {
                name: 'installHass',
                type: 'confirm',
                message: 'You need to install Home Assistant to continue, proceed?',
            },
        ])
        if (!hassInstallation['installHass']) {
            cli.stream.println(`Quitting`)
            cli.stream.hint('You can install Home Assistant by running ' + chalk.blue('sam hass install'))
            throw new Error(`Hass isn't installed, aborting`)
        }

        cli.stream.println('Installing Home Assistant, this can take a while')
        return await this.ssh.installHomeAssistant(
            _ => {},
            stepTitle => {
                cli.stream.println(chalk.green(stepTitle))
            })
            .then(async _ => {
                let loadingText = 'Waiting for Home Assistant initial setup. This could take a while.'
                cli.stream.loading(loadingText)
                let homeAssistantConfigurationDetected = false
                while (homeAssistantConfigurationDetected === false) {
                    await timeout(5000)
                    homeAssistantConfigurationDetected = await this.ssh.fileExists('/home/homeassistant/.homeassistant/configuration.yaml')
                    loadingText += ' . '
                    cli.stream.loading(loadingText)
                }
                cli.stream.success()
                return 'success'
            })
            .then(_ => this.hassCfgPath)
    }
}

